from frasco import Feature, action, current_app, request, abort, listens_to, current_context
from frasco.utils import (AttrDict, import_string, populate_obj, RequirementMissingError,\
                          find_classes_in_module, slugify)
from frasco.expression import compile_expr, eval_expr
from frasco.templating import FileLoader, FileSystemLoader
from werkzeug.local import LocalProxy
from .backend import *
from .utils import *
from .query import *
import inspect
import os
import inflection


_db = None

def get_current_db():
    return _db

db = LocalProxy(get_current_db)


form_imported = False
try:
    from .form import *
    form_imported = True
except ImportError:
    pass


class ModelsFeature(Feature):
    name = "models"
    defaults = {"backend": "persistpy",
                "pagination_per_page": 10,
                "scopes": {},
                "import_models": True,
                "ensure_schema": True,
                "admin_models": []}
    
    def init_app(self, app):
        self.backend_cls = self.get_backend_class(self.options["backend"])
        self.backend = self.backend_cls(app, self.options)
        self.scopes = compile_expr(self.options["scopes"])
        self.connected = False
        self.models = {}

        global _db
        self.db = _db = self.backend.db

        @listens_to("before_task")
        @listens_to("before_command")
        @app.before_request
        def before_request(*args, **kwargs):
            if not self.connected:
                self.backend.connect()
                self.connected = True
        @listens_to("after_task")
        @listens_to("after_command")
        @app.teardown_request
        def teardown_request(*args, **kwargs):
            if self.connected:
                self.backend.close()
                self.connected = False

        if self.options["import_models"]:
            models_pkg = self.options['import_models']
            if not isinstance(self.options['import_models'], str):
                models_pkg = "models"
            if app.import_name != "__main__":
                models_pkg = app.import_name + "." + models_pkg
            try:
                __import__(models_pkg)
            except ImportError:
                pass

        if form_imported:
            app.jinja_env.loader.bottom_loaders.append(FileLoader(
                os.path.join(os.path.dirname(__file__), "form_template.html"), "model_form_template.html"))
            app.jinja_env.loader.bottom_loaders.append(FileLoader(
                os.path.join(os.path.dirname(__file__), "bs_form_template.html"), "model_bs_form_template.html"))

    def init_admin(self, admin, app):
        from .admin import create_model_admin_blueprint
        app.jinja_env.loader.bottom_loaders.append(FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "admin/templates")))
        for model in self.options['admin_models']:
            kwargs = {}
            if isinstance(model, dict):
                model, kwargs = model.items()[0]
            model = self.ensure_model(model)
            with_counter = kwargs.pop('with_counter', False)
            counter_filters = kwargs.pop('counter_filters', {})
            title = inflection.pluralize(inflection.humanize(model.__name__))
            kwargs.setdefault('title', title)
            kwargs.setdefault('menu', title)
            name = inflection.pluralize(inflection.underscore(model.__name__))
            admin.register_blueprint(create_model_admin_blueprint(name, __name__, model, **kwargs))
            if with_counter:
                admin.register_dashboard_counter(title,
                    lambda: self.query(model).filter(**counter_filters).count(),
                    icon=kwargs.get('icon'))

    def get_backend_class(self, name):
        try:
            backend_cls = import_string("frasco_models.backends.%s" % name)
        except ImportError:
            backend_cls = import_string(name)

        if inspect.ismodule(backend_cls):
            # Gives the possibility to reference a module and auto-discover the Backend class
            classes = find_classes_in_module(backend_cls, (Backend,))
            if not classes:
                raise ImportError("Cannot find a Backend class in module '%s'" % name)
            if len(classes) > 1:
                raise ImportError("Model backend '%s' references a module with multiple backends" % name)
            backend_cls = classes[0]
        elif not issubclass(backend_cls, Backend):
            raise ImportError("Class '%s' is not a subclass of Backend" % name)

        return backend_cls

    def require_backend(self, name):
        if self.backend.name != name:
            raise RequirementMissingError("A models backend named '%s' is required but '%s' is used" % (name, self.backend.name))

    def ensure_model(self, model_name, **fields):
        if inspect.isclass(model_name):
            model_name = model_name.__name__
        if model_name not in self.models:
            self.models[model_name] = self.backend.ensure_model(model_name)
        if fields and self.options['ensure_schema']:
            for k, v in fields.iteritems():
                if not isinstance(v, dict):
                    fields[k] = dict(type=v)
            self.backend.ensure_schema(model_name, fields)
        return self.models[model_name]

    def __getitem__(self, name):
        return self.ensure_model(name)

    def __setitem__(self, name, model):
        self.models[name] = model

    def __contains__(self, name):
        return name in self.models

    def query(self, model):
        return Query(self.ensure_model(model), self.backend)

    def scoped_query(self, model, scope=None):
        q = self.query(model)
        if "model_scopes" in current_context.data:
            q = q.filter(**current_context.data.model_scopes.get(model.__name__, {}))
        if scope:
            scopes = scope if isinstance(scope, list) else list([scope])
            for s in scopes:
                if s not in self.scopes:
                    raise QueryError("Missing model scope '%s'" % s)
                q = q.filter(**eval_expr(self.scopes[s], current_context.vars))
        return q

    @action("build_model_query")
    def build_query(self, model, scope=None, filter_from=None, search_query=None, search_query_default_field=None,
                    order_by=None, limit=None, offset=None, **kwargs):
        q = self.scoped_query(model, scope)

        filters = {}
        if filter_from == "form":
            filters.update(dict([(f.name, f.data) for f in current_context.data.form]))
        elif filter_from == "url":
            filters.update(dict([(k, v) for k, v in request.values.items()]))
        elif filter_from == "args":
            filters.update(dict([(k, v) for k, v in request.view_args.items()]))
        if 'filters_or' in kwargs:
            q = q.filter(or_(*kwargs.pop('filters_or')))
        filters.update(kwargs.get("filters", kwargs))

        if filters:
            q = q.filter(**filters)
        if search_query:
            q = q.filter(*parse_search_query(search_query, search_query_default_field))
        if order_by:
            q = q.order_by(order_by)
        if limit:
            q = q.limit(limit)
        if offset:
            q = q.offset(offset)

        return q

    @action("paginate_query")
    def paginate(self, query, page=None, per_page=None, check_bounds=True):
        if page is None:
            page = int(page or request.values.get("page", 1))
        if per_page is None:
            per_page = self.options["pagination_per_page"]
        total = query.order_by(None).offset(None).limit(None).count()
        pagination = Pagination(page, per_page, total)
        if check_bounds and pagination.nb_pages > 0 and (page < 1 or page > pagination.nb_pages):
            raise PageOutOfBoundError()
        return query.offset(pagination.offset).limit(per_page), pagination

    @action("find_model")
    def find_first(self, model, not_found_404=True, **query):
        model = self.ensure_model(model)
        obj = self.build_query(model, **query).first()
        if obj is None and not_found_404:
            abort(404)
        if not self.find_first.as_:
            self.find_first.as_ = as_single_model(model)
        current_context.data.model = obj
        return obj

    @action("find_models", default_option="model")
    def find_all(self, model, paginate=False, page=None, pagination_var="pagination", **query):
        model = self.ensure_model(model)
        q = self.build_query(model, **query)

        if paginate:
            per_page = paginate if not isinstance(paginate, bool) else None
            try:
                q, pagination = self.paginate(q, page, per_page)
            except PageOutOfBoundError:
                abort(404)
            current_context.vars[pagination_var] = pagination

        if not self.find_all.as_:
            self.find_all.as_ = as_many_models(model)
        current_context.data.models = q
        return q

    @action("count_models", default_option="model")
    def count(self, model, **query):
        model = self.ensure_model(model)
        count = self.build_query(model, **query).count()
        if not self.count.as_:
            self.count.as_ = "%s_count" % as_single_model(model)
        return count

    @action("create_model", default_option="model")
    def create(self, model, **attrs):
        obj = self.ensure_model(model)(**clean_kwargs_proxy(attrs))
        if not self.create.as_:
            self.create.as_ = as_single_model(obj.__class__)
        return obj

    @action("save_model", default_option="obj")
    def save(self, obj=None, model=None, **attrs):
        auto_assign = False
        obj = clean_proxy(obj)
        if obj is None:
            obj = self.ensure_model(model)()
            auto_assign = True
        if attrs:
            populate_obj(obj, clean_kwargs_proxy(attrs))
        self.backend.save(obj)
        if not self.save.as_ and auto_assign:
            self.save.as_ = as_single_model(obj.__class__)
        return obj

    @action("create_model_from_form", default_option="model", requires=["form"])
    def create_from_form(self, model, form=None, **attrs):
        form = form or current_context.data.form
        obj = self.ensure_model(model)()
        form.populate_obj(obj)
        populate_obj(obj, clean_kwargs_proxy(attrs))
        if not self.create_from_form.as_:
            self.create_from_form.as_ = as_single_model(obj.__class__)
        return obj

    @action("save_form_model", default_option="model", requires=["form"])
    def save_from_form(self, obj=None, model=None, form=None, **attrs):
        form = form or current_context.data.form
        obj = clean_proxy(obj)
        auto_assign = False
        if obj is None:
            if isinstance(model, str):
                obj = self.ensure_model(model)()
                auto_assign = True
            else:
                obj = model()
        form.populate_obj(obj)
        populate_obj(obj, clean_kwargs_proxy(attrs))
        self.backend.save(obj)
        if not self.save_from_form.as_ and auto_assign:
            self.save_from_form.as_ = as_single_model(obj.__class__)
        return obj

    @action("delete_model", default_option="obj")
    def delete(self, obj):
        self.backend.remove(obj)

    @action("create_form_from_model", default_option="model", requires=["form"])
    def create_form_from_model(self, model, **kwargs):
        return create_form_from_model(model, **kwargs)

    @action("check_model_not_exists")
    def check_not_exists(self, model, error_message=None, **query):
        q = self.build_query(model, **query)
        if q.count() > 0:
            if error_message:
                flash(error_message, "error")
            current_context.exit(trigger_action_group="model_exists")

    @action("define_model_scope")
    def define_scope(self, model, **filters):
        current_context.data.setdefault("model_scopes", {})
        current_context.data.model_scopes.setdefault(model, {})
        current_context.data.model_scopes[model].update(filters.get('filters', filters))

    @action(as_="slug")
    def create_unique_slug(self, value, model, column="slug", **kwargs):
        slug = slugify(value)
        return ensure_unique_value(model, column, slug, **kwargs)


from frasco import current_app
from frasco_forms import TemplateForm
from frasco_forms.form import field_type_map
from frasco.utils import unknown_value
from wtforms.fields.core import UnboundField
import inflection
import datetime
import inspect


__all__ = ('create_form_from_model', 'create_form_class_from_model')


class ModelFormGenerationError(Exception):
    pass


model_type_map = dict([(str, "text"), (unicode, "text"), (int, "int"),
    (float, "float"), (datetime.datetime, "datetime5"), (datetime.date, "date5"),
    (bool, "checkbox")])


def create_form_from_model(model, **kwargs):
    obj = None
    if not inspect.isclass(model):
        obj = model
    return create_form_class_from_model(model, **kwargs)(obj=obj)


def create_form_class_from_model(model, backend=None, template=unknown_value, fields=None):
    if not backend:
        backend = current_app.features.models.backend
    if template is unknown_value:
        if 'bootstrap' in current_app.features:
            template = "model_bs_form_template.html"
        else:
            template = "model_form_template.html"

    model_name = model.__name__ if inspect.isclass(model) else model.__class__.__name__
    form_name = model_name + 'Form'
    form_class = type(form_name, (TemplateForm,), {"name": form_name, "template": template})

    names = []
    specs = {}
    inspected_fields = backend.inspect_fields(model)
    if fields:
        for f in fields:
            if isinstance(f, tuple):
                specs[f[0]] = f[1]
                names.append(f[0])
            else:
                names.append(f)
    else:
        names = [n for n, _ in inspected_fields]

    for name in names:
        spec = specs.get(name)
        if not spec:
            spec = dict(inspected_fields).get(name)
        if not spec:
            raise ModelFormGenerationError("Unknown field '%s'" % name)
        setattr(form_class, name, create_form_field_from_model_field(model, name, spec))

    return form_class


def create_form_field_from_model_field(model, name, spec):
    if isinstance(spec, UnboundField):
        return spec
    field_type = 'text'
    if spec.get('type'):
        field_type = model_type_map.get(spec['type'], "text")
    kwargs = {"label": spec.get('label', inflection.humanize(name))}
    if spec.get('description'):
        kwargs['description'] = spec['description']
    return field_type_map[field_type](**kwargs)


from frasco_admin import AdminBlueprint
from frasco import current_app, current_context, abort, request, redirect, url_for
from frasco_models.form import create_form_class_from_model
import inflection


def create_model_admin_blueprint(name, package, model, title=None, menu=None, icon=None, template_folder=None,
                                 list_columns=None, search_query_default_field=None, edit_actions=None,
                                 with_create=True, with_edit=True, with_delete=True, url_prefix=None,
                                 form_fields=None, filters=None, list_actions=None, can_edit=None, can_create=None):
    if not url_prefix:
        url_prefix = "/%s" % name
    bp = AdminBlueprint("admin_%s" % name, package, url_prefix=url_prefix,
            template_folder=template_folder)
    tpl_dir = name if template_folder else "models_default"

    def get_form_class():
        if not getattr(model, '__admin_form__', None):
            model.__admin_form__ = create_form_class_from_model(model,
                fields=getattr(model, '__admin_form_fields__', form_fields))
        return model.__admin_form__
    bp.get_form_class = get_form_class

    if hasattr(model, '__admin_search_query_default_field__'):
        search_query_default_field = model.__admin_search_query_default_field__
    if hasattr(model, '__admin_list_columns__'):
        list_columns = model.__admin_list_columns__
    if hasattr(model, '__admin_filters__'):
        filters = model.__admin_filters__

    if not edit_actions:
        edit_actions = []
    if not list_actions:
        list_actions = []
    if not filters:
        filters = {}
    if can_create is None:
        can_create = with_create
    if can_edit is None:
        can_edit = with_edit
    if with_delete:
        edit_actions.append(('Delete', '.delete', {'style': 'danger'}))

    @bp.view("/", template="admin/%s/index.html" % tpl_dir, admin_title=title, admin_menu=menu, admin_menu_icon=icon)
    def index():
        columns = list_columns
        if not columns:
            columns = []
            for name, _ in current_app.features.models.backend.inspect_fields(model):
                columns.append((name, inflection.humanize(name)))
        q = dict(order_by=request.args.get('sort', 'id'), **filters)
        s = request.args.get('search')
        if s:
            if s.startswith('#'):
                q['id'] = s[1:]
            else:
                q['search_query'] = s
                q['search_query_default_field'] = search_query_default_field
        current_context['actions'] = []
        if can_create:
            current_context['actions'].append(('Create', url_for('.create')))
        for label, url in list_actions:
            if callable(url):
                url = url()
            current_context['actions'].append((label, url))
        current_context['objs'] = current_app.features.models.find_all(model, paginate=15, **q)
        current_context['model_fields'] = [i[0] if isinstance(i, tuple) else i for i in columns]
        current_context['table_headers'] = [i[1] if isinstance(i, tuple) else inflection.humanize(i) for i in columns]
        current_context['can_create'] = can_create
        current_context['can_edit'] = can_edit

    if with_create:
        @bp.view("/create", template="admin/%s/create.html" % tpl_dir, methods=['GET', 'POST'])
        def create():
            form = get_form_class()()
            current_context['form'] = form
            if form.validate_on_submit():
                obj = current_app.features.models.save_from_form(model=model, form=form)
                return redirect(url_for('.edit', id=obj.id))

    if with_edit:
        @bp.view("/<id>", template="admin/%s/edit.html" % tpl_dir, methods=['GET', 'POST'])
        def edit(id):
            obj = current_app.features.models.query(model).get_or_404(id)
            form = get_form_class()(obj=obj)
            current_context['obj'] = obj
            current_context['admin_section_title'] = "Edit #%s" % obj.id
            current_context['form'] = form
            current_context['edit_actions'] = edit_actions
            if form.validate_on_submit():
                current_app.features.models.save_from_form(obj=obj, form=form)

    if with_delete:
        @bp.route("/<id>/delete")
        def delete(id):
            obj = current_app.features.models.query(model).get_or_404(id)
            current_app.features.models.backend.remove(obj)
            return redirect(url_for('.index'))

    return bp


from frasco import abort
import operator


def Q(**kwargs):
    return kwargs.items()


def and_(*args):
    return {"$and": args}


def or_(*args):
    return {"$or": args}


class QueryError(Exception):
    pass


class MultipleResultError(QueryError):
    pass


class NoResultError(QueryError):
    pass


class Query(object):
    ASC = "ASC"
    DESC = "DESC"

    def __init__(self, model, backend):
        self.model = model
        self.backend = backend
        self._fields = set()
        self._filters = []
        self._order_by = []
        self._offset = None
        self._limit = None

    def get(self, id):
        return self.backend.find_by_id(self.model, id)

    def get_or_404(self, id):
        obj = self.backend.find_by_id(self.model, id)
        if obj is None:
            abort(404)
        return obj

    def select(self, *args):
        return self.clone(_fields=args)

    def filter(self, *grouped_filters, **filters):
        q = self.clone()
        q._filters.extend(grouped_filters)
        q._filters.extend(filters.items())
        return q

    def order_by(self, field, direction=None):
        q = self.clone()
        if field is None:
            q._order_by = []
            return q
        if not isinstance(field, (list, tuple)):
            field = map(str.strip, field.split(','))
        for f in field:
            d = direction or self.ASC
            if isinstance(f, tuple):
                (f, d) = f
            elif " " in f:
                (f, d) = f.rsplit(" ", 1)
            q._order_by.append((f, d.upper()))
        return q

    def offset(self, offset):
        return self.clone(_offset=offset)

    def limit(self, limit):
        return self.clone(_limit=limit)

    def clone(self, **overrides):
        attr_to_clone = ('_fields', '_filters', '_order_by', '_offset', '_limit')
        data = {}
        for attr in attr_to_clone:
            v = getattr(self, attr)
            if isinstance(v, dict):
                data[attr] = dict(**v)
            elif isinstance(v, list):
                data[attr] = list(v)
            else:
                data[attr] = v
        data.update(overrides)
        q = self.__class__(self.model, self.backend)
        q.__dict__.update(data)
        return q

    def all(self):
        return self.backend.find_all(self)

    def first(self):
        return self.backend.find_first(self)

    def first_or_404(self):
        obj = self.first()
        if obj is None:
            abort(404)
        return obj

    def one(self):
        return self.backend.find_one(self)

    def count(self):
        return self.backend.count(self)

    def update(self, data):
        return self.backend.update(self, data)

    def delete(self):
        return self.backend.delete(self)

    def for_json(self):
        return {"model": self.model.__class__.__name__,
                "fields": self._fields,
                "filters": self._filters,
                "order_by": self._order_by,
                "offset": self._offset,
                "limit": self._limit}

    def __iter__(self):
        return iter(self.all())

    def __len__(self):
        return self.count()

    def __repr__(self):
        return "Query(fields=%s, filters=%s, order_by=%s, limit=%s, offset=%s)" %\
            (self._fields, self._filters, self._order_by, self._limit, self._offset)


known_operators = ('eq', 'ne', 'lt', 'lte', 'gt', 'gte', 'in', 'nin', 'contains',
                   'incr', 'push')

operators_mapping = {
    'eq': operator.eq,
    'ne': operator.ne,
    'lt': operator.lt,
    'lte': operator.le,
    'gt': operator.gt,
    'gte': operator.ge
}


def split_field_operator(field, check_operator=True, with_python_operator=False):
    operator = 'eq'
    if '__' in field:
        field, operator = field.split('__', 1)
    if check_operator and operator not in known_operators:
        raise QueryError("Unknown operator '%s'" % operator)
    if with_python_operator:
        return field, operator, operators_mapping.get(operator)
    return field, operator


from __future__ import absolute_import
from frasco import copy_extra_feature_options
from frasco.utils import JSONEncoder
from frasco_models import Backend, ModelSchemaError, and_, split_field_operator
from frasco_models.utils import clean_proxy
from flask.ext.mongoengine import (MongoEngine, Document as FlaskDocument,\
                                   DynamicDocument as FlaskDynamicDocument,\
                                   BaseQuerySet as FlaskQuerySet)
from mongoengine import Q, DynamicDocument as BaseDynamicDocument, ListField
from mongoengine.base import get_document, BaseDocument
from pymongo.read_preferences import ReadPreference
from bson import json_util
from bson.objectid import ObjectId


class MongoEngineJSONEncoder(JSONEncoder):
    """A JSONEncoder which provides serialization of MongoEngine documents
    """
    def default(self, obj):
        if isinstance(obj, BaseDocument) and not getattr(obj, 'for_json', None):
            return json_util._json_convert(obj.to_mongo())
        return JSONEncoder.default(self, obj)


class BaseQuerySet(FlaskQuerySet):
    """QuerySet with a for_json() method for easy encoding"""
    def for_json(self):
        return list(self.all())


class Document(FlaskDocument):
    meta = {'abstract': True,
            'queryset_class': BaseQuerySet}

    def __taskdump__(self):
        return 'frasco::current_app.features.models[%s]' % self.__class__.__name__, str(self.id)

    @classmethod
    def __taskload__(cls, id):
        return cls.objects.get(id=id)

    ######################################################
    #### BACKPORT FROM DEV VERSION
    ######################################################
    def reload(self, *fields, **kwargs):
        """Reloads all attributes from the database.
        :param fields: (optional) args list of fields to reload
        :param max_depth: (optional) depth of dereferencing to follow
        .. versionadded:: 0.1.2
        .. versionchanged:: 0.6  Now chainable
        .. versionchanged:: 0.9  Can provide specific fields to reload
        """
        max_depth = 1
        if fields and isinstance(fields[0], int):
            max_depth = fields[0]
            fields = fields[1:]
        elif "max_depth" in kwargs:
            max_depth = kwargs["max_depth"]

        if not self.pk:
            raise self.DoesNotExist("Document does not exist")
        obj = self._qs.read_preference(ReadPreference.PRIMARY).filter(
            **self._object_key).only(*fields).limit(1
                                                    ).select_related(max_depth=max_depth)

        if obj:
            obj = obj[0]
        else:
            raise self.DoesNotExist("Document does not exist")

        for field in self._fields_ordered:
            if not fields or field in fields:
                try:
                    setattr(self, field, self._reload(field, obj[field]))
                except KeyError:
                    # If field is removed from the database while the object
                    # is in memory, a reload would cause a KeyError
                    # i.e. obj.update(unset__field=1) followed by obj.reload()
                    delattr(self, field)

        # BUG FIX BY US HERE:
        if not fields:
            self._changed_fields = obj._changed_fields
        else:
            for field in fields:
                field = self._db_field_map.get(field, field)
                if field in self._changed_fields:
                    self._changed_fields.remove(field)
        self._created = False
        return self
    ######################################################
    ######################################################


class DynamicDocument(FlaskDynamicDocument):
    meta = {'abstract': True,
            'queryset_class': BaseQuerySet}


class MongoengineBackend(Backend):
    name = "mongoengine"

    def __init__(self, app, options):
        super(MongoengineBackend, self).__init__(app, options)
        copy_extra_feature_options(app.features.models, app.config, 'MONGODB_')
        self.db = MongoEngine(app)
        self.db.Document = Document
        self.db.DynamicDocument = DynamicDocument
        self.db.SetField = SetField
        # Flask-MongoEngine overrides the json_encoder but their
        # version ignores if for_json() is defined
        app.json_encoder = MongoEngineJSONEncoder

    def ensure_model(self, name):
        if isinstance(name, FlaskDocument):
            return name
        return get_document(name)

    def ensure_schema(self, name, fields):
        model = self.ensure_model(name)
        if isinstance(model, BaseDynamicDocument):
            return
        for fname, _ in fields.iteritems():
            if fname not in model._fields:
                raise ModelSchemaError("Missing field '%s' in model '%s'" % (fname, name))

    def find_by_id(self, model, id):
        if not isinstance(id, ObjectId):
            id = ObjectId(id)
        return model.objects.filter(id=id).first()

    def find_all(self, query):
        return self._transform_query(query).all()

    def find_first(self, query):
        return self._transform_query(query).first()

    def find_one(self, query):
        return self._transform_query(query).first()

    def count(self, query):
        return self._transform_query(query).count()

    def update(self, query, data):
        return self._transform_query(query).update(**self._prepare_data(data))

    def delete(self, query):
        return self._transform_query(query).delete()

    def _transform_query(self, q):
        qs = q.model.objects
        if q._filters:
            qs = qs(self._transform_query_filter_group(and_(*q._filters)))
        if q._order_by:
            qs = qs.order_by(*[''.join(('+' if v == "ASC" else '-', k)) for k, v in q._order_by])
        if q._offset:
            qs = qs.skip(q._offset)
        if q._limit:
            qs = qs.limit(q._limit)
        return qs

    def _transform_query_filter_group(self, group):
        operator, filters = group.items()[0]
        qs = None
        for filter in filters:
            if isinstance(filter, dict):
                q = self._transform_query_filter_group(filter)
                if q is None:
                    continue
            else:
                q = self._transform_query_filter(filter)
            if qs is None:
                qs = q
            elif operator == '$or':
                qs |= q
            else:
                qs &= q
        return qs

    def _transform_query_filter(self, filter):
        field, value = filter
        field, operator = split_field_operator(field)
        value = clean_proxy(value)
        if operator not in ('eq', 'contains'):
            field = '%s__%s' % (field, operator)
        return Q(**dict([(field, value)]))

    def _prepare_data(self, data):
        out = {}
        for field, value in data.iteritems():
            field, operator = split_field_operator(field)
            if operator == 'incr':
                out['inc__%s' % field] = value
            elif operator == 'push':
                out['push__%s' % field] = value
            else:
                out[field] = value
        return out


class SetField(ListField):
    """ Set field.

        Extends ListField, so that's how it's represented in Mongo.
    """
    def __set__(self, instance, value):
        return super(SetField, self).__set__(instance, set(value or []))

    def to_mongo(self, value):
        return super(SetField, self).to_mongo(list(value))

    def to_python(self, value):
        return set(super(SetField, self).to_python(value))

    def validate(self, value):
        if not isinstance(value, set):
            self.error('Only sets may be used.')


from __future__ import absolute_import
from frasco_models import Backend, and_, split_field_operator, QueryError
from persistpy import *
from ..utils import clean_proxy


class PersistpyBackend(Backend):
    name = "persistpy"

    def __init__(self, app, options):
        super(PersistpyBackend, self).__init__(app, options)
        self.client = MongoClient(options.get("url", "localhost"))
        self.session = Session(self.client[options["db"]])

    def make_model_base(self):
        return self.session.create_model_base()

    def ensure_model(self, name):
        return self.session.mapper.ensure_model(name)

    def ensure_schema(self, name, fields):
        model = self.session.mapper.ensure_model(name)
        for name, spec in fields.iteritems():
            type = spec.pop("type", None)
            if type in (list, dict) and "default" not in spec:
                spec["default"] = type
            model.ensure(**dict([(name, Field(**spec))]))

    def find_by_id(self, model, id):
        return model.query.get(id)

    def find_all(self, query):
        return self._transform_query(query).all()

    def find_first(self, query):
        return self._transform_query(query).first()

    def find_one(self, query):
        return self._transform_query(query).first()

    def count(self, query):
        return self._transform_query(query).count()

    def update(self, query, data):
        return self._transform_query(query).update(self._prepare_data(data))

    def delete(self, query):
        return self._transform_query(query).delete()

    def _transform_query(self, q):
        pq = q.model.query
        if q._filters:
            pq._spec = self._transform_query_filter_group(and_(*q._filters))
        if q._order_by:
            pq._sort = [(k, 1 if v == "ASC" else -1) for k, v in q._order_by]
        if q._offset:
            pq._skip = q._offset
        if q._limit:
            pq._limit = q._limit
        return pq

    def _transform_query_filter_group(self, group):
        operator, filters = group.popitem()
        conditions = []
        for filter in filters:
            if isinstance(filter, dict):
                conditions.append(self._transform_query_filter_group(filter))
            else:
                conditions.append(self._transform_query_filter(filter))
        return dict([(operator, conditions)])

    def _transform_query_filter(self, filter):
        field, value = filter
        field, operator = split_field_operator(field)
        value = clean_proxy(value)
        if field == 'id':
            field = '_id'
        if field == '_id' and not isinstance(value, ObjectId):
            value = ObjectId(value)
        if operator not in ('eq', 'contains'):
            value = dict([("$%s" % operator, value)])
        return dict([(field, value)])

    def _prepare_data(self, data):
        out = {}
        incr = {}
        push = {}
        for field, value in data.iteritems():
            field, operator = split_field_operator(field)
            if operator == 'incr':
                incr[field] = value
            elif operator == 'push':
                push[field] = value
            else:
                out[field] = value
        if incr:
            out['$inc'] = incr
        if push:
            out['$push'] = push
        return out


import inflection
from werkzeug import LocalProxy
import math
from flask import current_app
from .query import or_


def as_single_model(model):
    return inflection.underscore(model.__name__)


def as_many_models(model):
    return inflection.pluralize(as_single_model(model))


def clean_proxy(value):
    if isinstance(value, LocalProxy):
        return value._get_current_object()
    return value


def clean_kwargs_proxy(kwargs):
    for k, v in kwargs.iteritems():
        kwargs[k] = clean_proxy(v)
    return kwargs


class Pagination(object):
    def __init__(self, page, per_page, total):
        self.page = page
        self.per_page = per_page
        self.total = total
        self.nb_pages = int(math.ceil(float(self.total) / float(per_page)))

    @property
    def offset(self):
        return (self.page - 1) * self.per_page

    @property
    def prev_page(self):
        return self.page - 1 if self.page > 1 else None

    @property
    def next_page(self):
        return self.page + 1 if self.page < self.nb_pages else None

    @property
    def prev(self):
        if self.prev_page is None:
            return None
        return Pagination(self.prev_page, self.per_page, self.total)

    @property
    def next(self):
        if self.next_page is None:
            return None
        return Pagination(self.next_page, self.per_page, self.total)

    def iter_pages(self, left_edge=2, left_current=2,
                   right_current=5, right_edge=2):
        last = 0
        for num in xrange(1, self.nb_pages + 1):
            if num <= left_edge or \
               (num > self.page - left_current - 1 and \
                num < self.page + right_current) or \
               num > self.nb_pages - right_edge:
                if last + 1 != num:
                    yield None
                yield num
                last = num


class PageOutOfBoundError(Exception):
    pass


def move_obj_position_in_collection(obj, new_position, position_field='position', scope=None, data=None):
    if not data:
        data = {}
    current_position = getattr(obj, position_field, 0)
    up = new_position > current_position
    shift = -1 if up else 1
    lower_idx = min(current_position, new_position)
    lower_idx += 1 if up else 0
    upper_idx = max(current_position, new_position)
    upper_idx -= 0 if up else 1

    q = current_app.features.models.query(obj.__class__)
    if scope:
        q = q.filter(**scope)
    q = q.filter(**dict([('%s__gte' % position_field, lower_idx), ('%s__lte' % position_field, upper_idx)]))
    q.update(dict(dict([('%s__incr' % position_field, shift)]), **data))
    setattr(obj, position_field, new_position)


def ensure_unique_value(model, column, value, fallback=None, counter_start=1):
    if not fallback:
        fallback = value + "-%(counter)s"
    counter = counter_start
    q = current_app.features.models.query(model)
    while q.filter(**dict([(column, value)])).count() > 0:
        value = fallback % {'counter': counter}
        counter += 1
    return value


def parse_search_query(qs, default_field=None, default_op='AND'):
    parts = qs.split(' ')
    q = []
    default_field_values = []
    for part in parts:
        if ':' in part:
            field, value = part.split(':', 1)
            q.append((field, value))
        else:
            default_field_values.append(part)
    if default_field_values and default_field:
        if not isinstance(default_field, (list, tuple)):
            default_field = (default_field,)
        filters = []
        for field in default_field:
            filters.extend([(field, v) for v in default_field_values])
        q.append(or_(*filters))
    if default_op == 'OR':
        return or_(*q)
    return q



from .query import NoResultError
from frasco import AttrDict


class ModelNotFoundError(Exception):
    pass


class ModelSchemaError(Exception):
    pass


class RegisteringMetaClass(type):
    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        cls.__backend__.models[name] = cls


class Backend(object):
    def __init__(self, app, options):
        self.app = app
        self.options = options
        self.models = {}
        self._db = None

    @property
    def db(self):
        if not self._db:
            self._db = AttrDict(Model=self.make_model_base())
        return self._db

    @db.setter
    def db(self, db):
        self._db = db

    def make_model_base(self):
        return self.make_registering_model_base(object)

    def make_registering_model_base(self, base, name='Model'):
        return RegisteringMetaClass(name, (base,), {"__backend__": self})

    def connect(self):
        pass

    def close(self):
        pass

    def save(self, obj):
        obj.save()

    def remove(self, obj):
        obj.delete()

    def ensure_model(self, model_name):
        if model_name not in self.models:
            raise ModelNotFoundError('Model %s does not exist' % model_name)
        return self.models[model_name]

    def ensure_schema(self, model_name, fields):
        pass

    def inspect_fields(self, obj):
        fields = []
        for f in dir(obj):
            if not f.startswith('_'):
                fields.append((f, dict(type=None)))
        return fields

    def find_by_id(self, id):
        raise NotImplementedError()

    def find_all(self, query):
        raise NotImplementedError()

    def find_first(self, query):
        raise NotImplementedError()

    def find_one(self, query):
        obj = self.find_first(query)
        if not obj:
            raise NoResultError()
        return obj

    def count(self, query):
        raise NotImplementedError()

    def update(self, query, data):
        raise NotImplementedError()

    def delete(self, query):
        raise NotImplementedError()


from __future__ import absolute_import
from frasco import copy_extra_feature_options, current_app
from frasco.utils import JSONEncoder
from frasco_models import Backend, ModelSchemaError, and_, split_field_operator, QueryError
from frasco_models.utils import clean_proxy
from flask.ext.sqlalchemy import SQLAlchemy as BaseSQLAchemy, Model as BaseModel, _BoundDeclarativeMeta, _QueryProperty
from sqlalchemy.ext.declarative import declarative_base
import sqlalchemy
from sqlalchemy.inspection import inspect as sqlainspect
from sqlalchemy.sql import sqltypes
import inspect
import datetime


class Model(BaseModel):
    def __taskdump__(self):
        return 'frasco::current_app.features.models[%s]' % self.__class__.__name__, str(self.id)

    @classmethod
    def __taskload__(cls, id):
        return cls.query.get(id)


class SQLAlchemy(BaseSQLAchemy):
    def make_declarative_base(self, metadata=None):
        """Creates the declarative base."""
        base = declarative_base(cls=Model, name='Model',
                                metadata=metadata,
                                metaclass=_BoundDeclarativeMeta)
        base.query = _QueryProperty(self)
        return base


sqla_type_mapping = [
    (sqltypes.Integer, int),
    (sqltypes.Boolean, bool),
    (sqltypes.DateTime, datetime.datetime),
    (sqltypes.Date, datetime.date)
]


class SqlalchemyBackend(Backend):
    name = "sqlalchemy"

    def __init__(self, app, options):
        super(SqlalchemyBackend, self).__init__(app, options)
        copy_extra_feature_options(app.features.models, app.config, 'SQLALCHEMY_')
        self.db = SQLAlchemy(app)
        
        @app.cli.command()
        def create_db():
            try:
                self.db.create_all()
            except sqlalchemy.exc.CircularDependencyError as e:
                try:
                    self.graph_circular_dependency_error(e)
                except ImportError:
                    app.logger.info('Install networkx and pygraphviz to generate a graph of the circular dependency')
                    pass
                raise e

        app.cli.command('drop_db')(self.db.drop_all)

    def ensure_model(self, name):
        if isinstance(name, self.db.Model):
            return name
        return self.db.Model._decl_class_registry[name]

    def ensure_schema(self, name, fields):
        model = self.ensure_model(name)
        for fname, _ in fields.iteritems():
            if fname not in model.__mapper__.attrs:
                raise ModelSchemaError("Missing field '%s' in model '%s'" % (fname, name))

    def inspect_fields(self, model):
        if not inspect.isclass(model):
            model = model.__class__
        mapper = sqlainspect(model)
        fields = []
        for attr in mapper.column_attrs:
            field_type = str
            for coltype, pytype in sqla_type_mapping:
                if isinstance(attr.columns[0].type, coltype):
                    field_type = pytype
                    break
            fields.append((attr.key, dict(type=field_type)))
        return fields

    def save(self, obj):
        self.db.session.add(obj)
        self.db.session.commit()

    def remove(self, obj):
        self.db.session.delete(obj)
        self.db.session.commit()

    def find_by_id(self, model, id):
        return model.query.filter_by(id=id).first()

    def find_all(self, query):
        return self._transform_query(query).all()

    def find_first(self, query):
        return self._transform_query(query).first()

    def find_one(self, query):
        return self._transform_query(query).first()

    def count(self, query):
        return self._transform_query(query).count()

    def update(self, query, data):
        return self._transform_query(query).update(
            self._prepare_data(query.model, data),
            synchronize_session=False)

    def delete(self, query):
        return self._transform_query(query).delete(
            synchronize_session=False)

    def _transform_query(self, q):
        qs = q.model.query
        if q._filters:
            qs = qs.filter(self._transform_query_filter_group(q.model, and_(*q._filters)))
        if q._order_by:
            qs = qs.order_by(*[k + ' ' + v for k, v in q._order_by])
        if q._offset:
            qs = qs.offset(q._offset)
        if q._limit:
            qs = qs.limit(q._limit)
        return qs

    def _transform_query_filter_group(self, model, group):
        operator, filters = group.items()[0]
        transformed_filters = []
        for filter in filters:
            if isinstance(filter, dict):
                q = self._transform_query_filter_group(model, filter)
                if q is None:
                    continue
            else:
                q = self._transform_query_filter(model, filter)
            transformed_filters.append(q)
        if operator == "$or":
            return sqlalchemy.or_(*transformed_filters)
        return sqlalchemy.and_(*transformed_filters)

    def _transform_query_filter(self, model, filter):
        field, value = filter
        field, operator, py_operator = split_field_operator(field, with_python_operator=True)
        value = clean_proxy(value)
        column = getattr(model, field)
        if py_operator:
            return py_operator(column, value)
        if operator == 'in':
            return column.in_(value)
        if operator == 'nin':
            return ~column.in_(value)
        raise QueryError("Cannot convert operator '%s' to sqlalchemy operator" % operator)

    def _prepare_data(self, model, data):
        out = {}
        for field, value in data.iteritems():
            field, operator = split_field_operator(field)
            column = getattr(model, field)
            if operator == 'incr':
                out[column] = column + value
            elif operator == 'push':
                raise QueryError("Operator 'push' not supported by sqlalchemy")
            else:
                out[column] = value
        return out

    def graph_circular_dependency_error(self, e, filename='sqla_circular_dep_graph.png'):
        # from: http://ilyasterin.com/blog/2014/01/cyclical-dependency-detection-in-the-database.html
        import networkx as nx
        G=nx.DiGraph()
        cycle_tables = set([t.name for t in e.cycles])
        for t in e.cycles:
            for fk in t.foreign_keys:
                table, col = fk.target_fullname.split('.')
                if (table in cycle_tables):
                    G.add_edge(t.name, table)
        agraph = nx.to_agraph(G)
        agraph.draw(filename, format='png', prog='dot')